"""Functions to create integral moments."""


def make_integral_moment_dofs(
    reference,
    vertices=None, edges=None, faces=None, volumes=None,
    cells=None, facets=None, ridges=None, peaks=None,
    variant="equispaced"
):
    """Generate DOFs due to integral moments on sub entities.

    Parameters
    ----------
    reference: symfem.references.Reference
        The reference cell.
    vertices: tuple
        DOFs on dimension 0 entities.
    edges: tuple
        DOFs on dimension 1 entities.
    faces: tuple
        DOFs on dimension 2 entities.
    volumes: tuple
        DOFs on dimension 3 entities.
    cells: tuple
        DOFs on codimension 0 entities.
    facets: tuple
        DOFs on codimension 1 entities.
    ridges: tuple
        DOFs on codimension 2 entities.
    peaks: tuple
        DOFs on codimension 3 entities.
    """
    from symfem import create_reference
    dofs = []

    # DOFs per dimension
    for dim, moment_data in enumerate([vertices, edges, faces, volumes]):
        if moment_data is not None:
            IntegralMoment, SubElement, order = moment_data
            if order >= SubElement.min_order:
                sub_type = reference.sub_entity_types[dim]
                if sub_type is not None:
                    assert dim > 0
                    for i, vs in enumerate(reference.sub_entities(dim)):
                        sub_ref = create_reference(
                            sub_type,
                            vertices=[reference.reference_vertices[v] for v in vs])
                        sub_element = SubElement(sub_ref, order, variant=variant)
                        for f, d in zip(
                            sub_element.get_basis_functions(), sub_element.dofs
                        ):
                            dofs.append(IntegralMoment(sub_ref, f, d, entity=(dim, i)))

    # DOFs per codimension
    for _dim, moment_data in enumerate([peaks, ridges, facets, cells]):
        dim = reference.tdim - 3 + _dim
        if moment_data is not None:
            IntegralMoment, SubElement, order = moment_data
            if order >= SubElement.min_order:
                sub_type = reference.sub_entity_types[dim]
                if sub_type is not None:
                    assert dim > 0
                    for i, vs in enumerate(reference.sub_entities(dim)):
                        sub_ref = create_reference(
                            sub_type,
                            vertices=[reference.reference_vertices[v] for v in vs])
                        sub_element = SubElement(sub_ref, order, variant=variant)
                        for f, d in zip(
                            sub_element.get_basis_functions(), sub_element.dofs
                        ):
                            dofs.append(IntegralMoment(sub_ref, f, d, entity=(dim, i)))

    return dofs



import sympy
from symfem import create_element


def all_equal(a, b):
    if isinstance(a, (list, tuple)):
        for i, j in zip(a, b):
            if not all_equal(i, j):
                return False
        return True
    return a == b


def test_lagrange():
    space = create_element("triangle", "Lagrange", 1)
    assert all_equal(
        space.tabulate_basis([[0, 0], [0, 1], [1, 0]]),
        ((1, 0, 0), (0, 0, 1), (0, 1, 0)),
    )


def test_nedelec():
    space = create_element("triangle", "Nedelec", 1)
    assert all_equal(
        space.tabulate_basis([[0, 0], [1, 0], [0, 1]], "xxyyzz"),
        ((0, 0, 1, 0, 1, 0), (0, 0, 1, 1, 0, 1), (-1, 1, 0, 0, 1, 0)),
    )


def test_rt():
    space = create_element("triangle", "Raviart-Thomas", 1)
    assert all_equal(
        space.tabulate_basis([[0, 0], [1, 0], [0, 1]], "xxyyzz"),
        ((0, -1, 0, 0, 0, 1), (-1, 0, -1, 0, 0, 1), (0, -1, 0, -1, 1, 0)),
    )


def test_Q():
    space = create_element("quadrilateral", "Q", 1)
    assert all_equal(
        space.tabulate_basis([[0, 0], [1, 0], [0, 1], [1, 1]]),
        ((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1)),
    )


def test_dual0():
    space = create_element("dual polygon(4)", "dual", 0)
    q = sympy.Rational(1, 4)
    assert all_equal(
        space.tabulate_basis([[q, q], [-q, q], [-q, -q], [q, -q]]),
        ((1, ), (1, ), (1, ), (1, ))
    )


def test_dual1():
    space = create_element("dual polygon(4)", "dual", 1)
    h = sympy.Rational(1, 2)
    q = sympy.Rational(1, 4)
    e = sympy.Rational(1, 8)
    assert all_equal(
        space.tabulate_basis([[0, 0], [q, q], [h, 0]]),
        ((q, q, q, q),
         (sympy.Rational(5, 8), e, e, e),
         (sympy.Rational(3, 8), e, e, sympy.Rational(3, 8)))
    )



"""Nedelec elements on simplices.

These elements' definitions appear in https://doi.org/10.1007/BF01396415
(Nedelec, 1980) and https://doi.org/10.1007/BF01389668 (Nedelec, 1986)
"""

from ..core.finite_element import FiniteElement
from ..core.moments import make_integral_moment_dofs
from ..core.polynomials import polynomial_set, Hcurl_polynomials
from ..core.functionals import TangentIntegralMoment, IntegralMoment
from .lagrange import DiscontinuousLagrange, VectorDiscontinuousLagrange
from .rt import RaviartThomas


class NedelecFirstKind(FiniteElement):
    """Nedelec first kind Hcurl finite element."""

    def __init__(self, reference, order, variant):
        poly = polynomial_set(reference.tdim, reference.tdim, order - 1)
        poly += Hcurl_polynomials(reference.tdim, reference.tdim, order)
        dofs = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DiscontinuousLagrange, order - 1),
            faces=(IntegralMoment, VectorDiscontinuousLagrange, order - 2),
            volumes=(IntegralMoment, VectorDiscontinuousLagrange, order - 3),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["Nedelec", "Nedelec1", "N1curl"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    mapping = "covariant"
    continuity = "H(curl)"


class NedelecSecondKind(FiniteElement):
    """Nedelec second kind Hcurl finite element."""

    def __init__(self, reference, order, variant):
        poly = polynomial_set(reference.tdim, reference.tdim, order)

        dofs = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DiscontinuousLagrange, order),
            faces=(IntegralMoment, RaviartThomas, order - 1),
            volumes=(IntegralMoment, RaviartThomas, order - 2),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["Nedelec2", "N2curl"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    mapping = "covariant"
    continuity = "H(curl)"



"""Bernstein elements on simplices.

This element's definition appears in https://doi.org/10.1007/s00211-010-0327-2
(Kirby, 2011)
"""

from ..core.symbolic import one, x
from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import IntegralAgainst


def single_choose(n, k):
    """Calculate choose function of a set of powers."""
    out = one
    for i in range(k + 1, n + 1):
        out *= i
    for i in range(1, n - k + 1):
        out /= i
    return out


def choose(n, powers):
    """Calculate choose function of a set of powers."""
    out = one
    for p in powers:
        out *= single_choose(n, p)
        n -= p
    return out


def bernstein_polynomials(n, d):
    """
    Return a list of Bernstein polynomials.

    Parameters
    ----------
    n : int
        The polynomial order
    d : int
        The topological dimension
    """
    poly = []
    if d == 1:
        lambdas = [1 - x[0], x[0]]
        powers = [[i, n - i] for i in range(n + 1)]
    elif d == 2:
        lambdas = [1 - x[0] - x[1], x[0], x[1]]
        powers = [[i, j, n - i - j]
                  for i in range(n + 1)
                  for j in range(n + 1 - i)]
    elif d == 3:
        lambdas = [1 - x[0] - x[1] - x[2], x[0], x[1], x[2]]
        powers = [[i, j, k, n - i - j - k]
                  for i in range(n + 1)
                  for j in range(n + 1 - i)
                  for k in range(n + 1 - i - j)]

    for p in powers:
        f = choose(n, p)
        for a, b in zip(lambdas, p):
            f *= a ** b
        poly.append(f)

    return poly


class Bernstein(FiniteElement):
    """Bernstein finite element."""

    def __init__(self, reference, order, variant):

        dofs = [
            IntegralAgainst(reference, p, entity=(reference.tdim, 0))
            for p in bernstein_polynomials(order, reference.tdim)
        ]
        poly = polynomial_set(reference.tdim, 1, order)
        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Bernstein", "Bernstein-Bezier"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 0
    continuity = "C0"
    mapping = "identity"



"""Bell elements on triangle.

This element's definition is given in https://doi.org/10.1002/nme.1620010108 (Bell, 1969)
"""

from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import (PointEvaluation, PointNormalDerivativeEvaluation,
                                DerivativePointEvaluation)


class Bell(FiniteElement):
    """Bell finite element."""

    def __init__(self, reference, order, variant):
        assert reference.name == "triangle"
        assert order == 5
        from symfem import create_reference
        dofs = []
        for v_n, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(v, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(v, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(v, (0, 1), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(v, (2, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(v, (1, 1), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(v, (0, 2), entity=(0, v_n)))
        for e_n, e in enumerate(reference.edges):
            sub_ref = create_reference(
                reference.sub_entity_types[1],
                vertices=tuple(reference.vertices[i] for i in e))
            dofs.append(PointNormalDerivativeEvaluation(
                tuple((i + j) / 2 for i, j in zip(reference.vertices[e[0]],
                                                  reference.vertices[e[1]])),
                sub_ref, (1, e_n)))

        super().__init__(
            reference, order, polynomial_set(reference.tdim, 1, order), dofs, reference.tdim, 1
        )

    names = ["Bell"]
    references = ["triangle"]
    min_order = 5
    max_order = 5
    continuity = "C1"
    mapping = "identity"



"""Polynomial sets."""
from .symbolic import x, zero, one
from itertools import product


def polynomial_set_1d(dim, order):
    """One dimensional polynomial set."""
    if dim == 1:
        return [x[0] ** i for i in range(order + 1)]
    if dim == 2:
        return [
            x[0] ** i * x[1] ** j
            for j in range(order + 1)
            for i in range(order + 1 - j)
        ]
    if dim == 3:
        return [
            x[0] ** i * x[1] ** j * x[2] ** k
            for k in range(order + 1)
            for j in range(order + 1 - k)
            for i in range(order + 1 - k - j)
        ]


def polynomial_set(domain_dim, range_dim, order):
    """Polynomial set."""
    if range_dim == 1:
        return polynomial_set_1d(domain_dim, order)
    set1d = polynomial_set_1d(domain_dim, order)
    return [
        tuple(p if i == j else zero for j in range(range_dim))
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_polynomials(domain_dim, range_dim, order):
    """Hdiv conforming polynomial set."""
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            (
                x[0] * x[0] ** (order - 1 - j) * x[1] ** j,
                x[1] * x[0] ** (order - 1 - j) * x[1] ** j,
            )
            for j in range(order)
        ]
    if domain_dim == 3:
        basis = []
        for j in range(order):
            for k in range(order - j):
                p = x[0] ** (order - 1 - j - k) * x[1] ** j * x[2] ** k
                basis.append((x[0] * p, x[1] * p, x[2] * p))
        return basis


def Hcurl_polynomials(domain_dim, range_dim, order):
    """Hcurl conforming polynomial set."""
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            (
                x[0] ** (order - 1 - j) * x[1] ** (j + 1),
                -x[0] ** (order - j) * x[1] ** j,
            )
            for j in range(order)
        ]
    if domain_dim == 3:
        poly = []
        poly += [[x[0] ** (m - 1) * x[1] ** n * x[2] ** (order - m - n + 1),
                  zero,
                  -x[0] ** m * x[1] ** n * x[2] ** (order - m - n)]
                 for n in range(order) for m in range(1, order + 1 - n)]
        poly += [[zero,
                  x[0] ** m * x[1] ** (n - 1) * x[2] ** (order - m - n + 1),
                  -x[0] ** m * x[1] ** n * x[2] ** (order - m - n)]
                 for m in range(order) for n in range(1, order + 1 - m)]
        poly += [[x[0] ** (order - n) * x[1] ** n,
                  -x[0] ** (order + 1 - n) * x[1] ** (n - 1),
                  zero]
                 for n in range(1, order + 1)]
        return poly


def quolynomial_set_1d(dim, order):
    """One dimensional quolynomial set."""
    basis = []
    for j in product(range(order + 1), repeat=dim):
        poly = 1
        for a, b in zip(x, j):
            poly *= a ** b
        basis.append(poly)
    return basis


def quolynomial_set(domain_dim, range_dim, order):
    """Quolynomial set."""
    if range_dim == 1:
        return quolynomial_set_1d(domain_dim, order)
    set1d = quolynomial_set_1d(domain_dim, order)
    return [
        tuple(p if i == j else zero for j in range(range_dim))
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_quolynomials(domain_dim, range_dim, order):
    """Hdiv conforming quolynomial set."""
    assert domain_dim == range_dim
    basis = []
    for d in range(domain_dim):
        for j in product(range(order), repeat=domain_dim - 1):
            poly = one
            for a, b in zip(x, j[:d] + (order,) + j[d:]):
                poly *= a ** b
            basis.append(tuple(poly if i == d else zero for i in range(domain_dim)))
    return basis


def Hcurl_quolynomials(domain_dim, range_dim, order):
    """Hcurl conforming quolynomial set."""
    assert domain_dim == range_dim
    basis = []
    for d in range(domain_dim):
        for j in product(
            *[range(order) if i == d else range(order + 1) for i in range(domain_dim)]
        ):
            if order not in j:
                continue
            poly = one
            for a, b in zip(x, j):
                poly *= a ** b
            basis.append(tuple(poly if i == d else zero for i in range(domain_dim)))
    return basis


def serendipity_indices(total, linear, dim, done=[]):
    """Get the set indices for a serendipity polynomial set."""
    if len(done) == dim:
        if done.count(1) >= linear:
            return [done]
        return []
    if len(done) == dim - 1:
        return serendipity_indices(total, linear, dim, done=done + [total - sum(done)])
    out = []
    for i in range(total - sum(done) + 1):
        out += serendipity_indices(total, linear, dim, done + [i])
    return out


def serendipity_set_1d(dim, order):
    """One dimensional serendipity set."""
    basis = []
    for s in range(order + 1, order + dim + 1):
        for i in serendipity_indices(s, s - order, dim):
            p = one
            for j, k in zip(x, i):
                p *= j ** k
            basis.append(p)
    return basis


def serendipity_set(domain_dim, range_dim, order):
    """Serendipity set."""
    if range_dim == 1:
        return serendipity_set_1d(domain_dim, order)
    set1d = serendipity_set_1d(domain_dim, order)
    return [
        tuple(p if i == j else zero for j in range(range_dim))
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_serendipity(domain_dim, range_dim, order):
    """Hdiv conforming serendipity set."""
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            (x[0] ** (order + 1), (order + 1) * x[0] ** order * x[1]),
            ((order + 1) * x[0] * x[1] ** order, x[1] ** (order + 1)),
        ]
    if domain_dim == 3:
        a = []
        if order == 0:
            a.append((zero, x[0] * x[2], -x[0] * x[1]))
            a.append((x[1] * x[2], zero, -x[0] * x[1]))
        elif order == 1:
            for p in [x[1], x[2]]:
                a.append((zero, x[0] * x[2] * p, -x[0] * x[1] * p))
            for p in [x[0], x[2]]:
                a.append((x[1] * x[2] * p, zero, -x[0] * x[1] * p))
            for p in [x[0], x[1]]:
                a.append((x[1] * x[2] * p, -x[0] * x[2] * p, zero))
        elif order == 2:
            for p in [x[0] ** 2, x[0] * x[2], x[2] ** 2]:
                a.append((x[1] * x[2] * p, zero, -x[0] * x[1] * p))
            for p in [x[1] ** 2, x[1] * x[2], x[2] ** 2]:
                a.append((zero, x[0] * x[2] * p, -x[0] * x[1] * p))
            for p in [x[0] ** 2, x[0] * x[1], x[1] ** 2]:
                a.append((x[1] * x[2] * p, -x[0] * x[2] * p, zero))
        elif order == 3:
            for p in [x[0] ** 3, x[0] ** 2 * x[2], x[0] * x[2] ** 2, x[2] ** 3]:
                a.append((x[1] * x[2] * p, zero, -x[0] * x[1] * p))
            for p in [x[1] ** 3, x[1] ** 2 * x[2], x[1] * x[2] ** 2, x[1] ** 3]:
                a.append((zero, x[0] * x[2] * p, -x[0] * x[1] * p))
            for p in [x[0] ** 3, x[0] ** 2 * x[1], x[0] * x[1] ** 2, x[1] ** 3]:
                a.append((x[1] * x[2] * p, -x[0] * x[2] * p, zero))
        else:
            raise NotImplementedError
        return [
            (
                r.diff(x[1]) - q.diff(x[2]),
                p.diff(x[2]) - r.diff(x[0]),
                q.diff(x[0]) - p.diff(x[1]),
            )
            for p, q, r in a
        ]
    raise NotImplementedError


def Hcurl_serendipity(domain_dim, range_dim, order):
    """Hcurl conforming serendipity set."""
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            ((order + 1) * x[0] ** order * x[1], -x[0] ** (order + 1)),
            (x[1] ** (order + 1), -(order + 1) * x[0] * x[1] ** order),
        ]
    if domain_dim == 3:
        out = []
        if order == 1:
            out += [
                (zero, x[0] * x[2], -x[0] * x[1]),
                (x[1] * x[2], zero, -x[0] * x[1]),
            ]
        elif order == 2:
            for p in [x[0], x[2]]:
                out.append((x[1] * x[2] * p, zero, -x[0] * x[1] * p))
            for p in [x[1], x[2]]:
                out.append((zero, x[0] * x[2] * p, -x[0] * x[1] * p))
            for p in [x[0], x[1]]:
                out.append((x[1] * x[2] * p, -x[0] * x[2] * p, zero))
        elif order == 3:
            for p in [x[0] ** 2, x[0] * x[2], x[2] ** 2]:
                out.append((x[1] * x[2] * p, zero, -x[0] * x[1] * p))
            for p in [x[1] ** 2, x[1] * x[2], x[2] ** 2]:
                out.append((zero, x[0] * x[2] * p, -x[0] * x[1] * p))
            for p in [x[0] ** 2, x[0] * x[1], x[1] ** 2]:
                out.append((x[1] * x[2] * p, -x[0] * x[2] * p, zero))
        else:
            raise NotImplementedError
        for p in serendipity_set(domain_dim, 1, order + 1):
            out.append(tuple(p.diff(i) for i in x))
        return out

    raise NotImplementedError



"""Reduced Hsieh-Clough-Tocher elements on simplices.

This element's definition appears in https://doi.org/10.2307/2006147
(Ciarlet, 1978)
"""

import sympy
from ..core.finite_element import FiniteElement
from ..core.functionals import PointEvaluation, DerivativePointEvaluation
from ..core.polynomials import polynomial_set
from ..core.symbolic import sym_sum, PiecewiseFunction, zero, x, subs, one


class P1Hermite(FiniteElement):
    """P1Hermite finite element."""

    def __init__(self, reference, order, variant, poly):
        assert order == 3
        dofs = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(vs, (0, 1), entity=(0, v_n)))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    def perform_mapping(self, basis, map, inverse_map):
        """Map the basis onto a cell using the appropriate mapping for the element."""
        out = []
        tdim = self.reference.tdim
        J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
        for v in range(tdim + 1):
            out.append(basis[(tdim + 1) * v])
            for i in range(tdim):
                out.append(sym_sum(a * b for a, b in
                                   zip(basis[(tdim + 1) * v + 1: (tdim + 1) * (v + 1)],
                                       J.row(i))))
        if tdim == 2:
            out.append(basis[-1])
        if tdim == 3:
            out += basis[-4:]
        assert len(out) == len(basis)
        return [subs(b, x, inverse_map) for b in out]

    names = []
    references = ["triangle"]
    min_order = 3
    max_order = 3
    continuity = "C0"


class ReducedHsiehCloughTocher(FiniteElement):
    """Reduced Hsieh-Clough-Tocher finite element."""

    def __init__(self, reference, order, variant):
        from symfem import create_reference
        assert order == 3
        assert reference.name == "triangle"
        dofs = []
        for v_n, vs in enumerate(reference.sub_entities(0)):
            dofs.append(PointEvaluation(vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(vs, (0, 1), entity=(0, v_n)))

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid)]

        refs = [create_reference("triangle", vs) for vs in subs]

        polys = [
            polynomial_set(reference.tdim, 1, order),
            [],
            polynomial_set(reference.tdim, 1, order),
        ]
        polys[0].remove(x[0] ** 2 * x[1])
        polys[1] = [one, x[0], x[0] ** 2, x[1], x[0] * x[1], x[1] ** 2,
                    x[0] * x[1] ** 2 - x[0] ** 2 * x[1],
                    x[0] ** 3 - x[1] ** 3, x[0] ** 3 + 3 * x[0] * x[1] ** 2]
        polys[2].remove(x[0] * x[1] ** 2)

        bases = [
            P1Hermite(r, 3, variant, p).get_basis_functions()
            for r, p in zip(refs, polys)
        ]

        piece_list = []
        piece_list.append((bases[0][0], zero, bases[2][3]))
        piece_list.append((bases[0][1], zero, bases[2][4]))
        piece_list.append((bases[0][2], zero, bases[2][5]))
        piece_list.append((bases[0][3], bases[1][0], zero))
        piece_list.append((bases[0][4], bases[1][1], zero))
        piece_list.append((bases[0][5], bases[1][2], zero))
        # TODO: are these right to remove??
        # piece_list.append((bases[0][6], bases[1][6], bases[2][6]))
        # piece_list.append((bases[0][7], bases[1][7], bases[2][7]))
        # piece_list.append((bases[0][8], bases[1][8], bases[2][8]))
        piece_list.append((zero, bases[1][3], bases[2][0]))
        piece_list.append((zero, bases[1][4], bases[2][1]))
        piece_list.append((zero, bases[1][5], bases[2][2]))

        poly = [
            PiecewiseFunction(list(zip(subs, p)))
            for p in piece_list
        ]

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    def perform_mapping(self, basis, map, inverse_map):
        """Map the basis onto a cell using the appropriate mapping for the element."""
        out = []
        tdim = self.reference.tdim
        J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
        for v in range(tdim + 1):
            out.append(basis[(tdim + 1) * v])
            for i in range(tdim):
                out.append(sym_sum(a * b for a, b in
                                   zip(basis[(tdim + 1) * v + 1: (tdim + 1) * (v + 1)],
                                       J.row(i))))
        assert len(out) == len(basis)
        return [subs(b, x, inverse_map) for b in out]

    names = ["reduced Hsieh-Clough-Tocher", "rHCT"]
    references = ["triangle"]
    min_order = 3
    max_order = 3
    mapping = "identity"
    continuity = "C1"



"""Q elements on tensor product cells."""

import sympy
from itertools import product
from ..core.symbolic import one, zero
from ..core.finite_element import FiniteElement
from ..core.moments import make_integral_moment_dofs
from ..core.polynomials import quolynomial_set, Hdiv_quolynomials, Hcurl_quolynomials
from ..core.quadrature import get_quadrature
from ..core.functionals import (PointEvaluation, DotPointEvaluation, IntegralMoment,
                                TangentIntegralMoment, NormalIntegralMoment)


class Q(FiniteElement):
    """A Q element."""

    def __init__(self, reference, order, variant):
        from symfem import create_reference

        if order == 0:
            dofs = [
                PointEvaluation(
                    tuple(sympy.Rational(1, 2) for i in range(reference.tdim)),
                    entity=(reference.tdim, 0))]
        else:
            points, _ = get_quadrature(variant, order + 1)

            dofs = []
            for v_n, v in enumerate(reference.reference_vertices):
                dofs.append(PointEvaluation(v, entity=(0, v_n)))
            for edim in range(1, 4):
                for e_n, vs in enumerate(reference.sub_entities(edim)):
                    entity = create_reference(
                        reference.sub_entity_types[edim],
                        vertices=tuple(reference.reference_vertices[i] for i in vs))
                    for i in product(range(1, order), repeat=edim):
                        dofs.append(
                            PointEvaluation(
                                tuple(o + sum(a[j] * points[b]
                                              for a, b in zip(entity.axes, i[::-1]))
                                      for j, o in enumerate(entity.origin)),
                                entity=(edim, e_n)))

        super().__init__(
            reference, order,
            quolynomial_set(reference.tdim, 1, order),
            dofs,
            reference.tdim,
            1)

    names = ["Q"]
    references = ["interval", "quadrilateral", "hexahedron"]
    min_order = 0
    mapping = "identity"
    continuity = "C0"


class DiscontinuousQ(FiniteElement):
    """A dQ element."""

    def __init__(self, reference, order, variant):
        if order == 0:
            dofs = [
                PointEvaluation(
                    tuple(sympy.Rational(1, 2) for i in range(reference.tdim)),
                    entity=(reference.tdim, 0))]
        else:
            points, _ = get_quadrature(variant, order + 1)

            dofs = []
            for i in product(range(order + 1), repeat=reference.tdim):
                dofs.append(PointEvaluation(tuple(points[j] for j in i[::-1]),
                                            entity=(reference.tdim, 0)))

        super().__init__(
            reference, order,
            quolynomial_set(reference.tdim, 1, order),
            dofs,
            reference.tdim,
            1)

    names = ["dQ"]
    references = ["interval", "quadrilateral", "hexahedron"]
    min_order = 0
    mapping = "identity"
    continuity = "L2"


class VectorQ(FiniteElement):
    """A vector Q element."""

    def __init__(self, reference, order, variant):
        scalar_space = Q(reference, order, variant)
        dofs = []
        if reference.tdim == 1:
            directions = [1]
        else:
            directions = [
                tuple(one if i == j else zero for j in range(reference.tdim))
                for i in range(reference.tdim)
            ]
        for p in scalar_space.dofs:
            for d in directions:
                dofs.append(DotPointEvaluation(p.point, d, entity=p.entity))

        super().__init__(
            reference, order,
            quolynomial_set(reference.tdim, reference.tdim, order),
            dofs,
            reference.tdim,
            reference.tdim,
        )

    names = ["vector Q", "vQ"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    mapping = "identity"
    continuity = "C0"


class Nedelec(FiniteElement):
    """Nedelec Hcurl finite element."""

    def __init__(self, reference, order, variant):
        poly = quolynomial_set(reference.tdim, reference.tdim, order - 1)
        poly += Hcurl_quolynomials(reference.tdim, reference.tdim, order)

        dofs = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DiscontinuousQ, order - 1),
            faces=(IntegralMoment, RaviartThomas, order - 1),
            volumes=(IntegralMoment, RaviartThomas, order - 1),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["NCE", "RTCE", "Qcurl"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    mapping = "covariant"
    continuity = "H(curl)"


class RaviartThomas(FiniteElement):
    """Raviart-Thomas Hdiv finite element."""

    def __init__(self, reference, order, variant):
        poly = quolynomial_set(reference.tdim, reference.tdim, order - 1)
        poly += Hdiv_quolynomials(reference.tdim, reference.tdim, order)

        dofs = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DiscontinuousQ, order - 1),
            cells=(IntegralMoment, Nedelec, order - 1),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["NCF", "RTCF", "Qdiv"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    mapping = "contravariant"
    continuity = "H(div)"



"""Functions to map functions between cells."""

import sympy
from .symbolic import subs, x
from .vectors import vdot


def identity(f, map, inverse_map, tdim):
    """Map functions."""
    g = subs(f, x, inverse_map)
    return g


def covariant(f, map, inverse_map, tdim):
    """Map H(curl) functions."""
    g = subs(f, x, inverse_map)
    J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
    Jinv = J.inv().transpose()
    return tuple(vdot(Jinv.row(i), g) for i in range(Jinv.rows))


def contravariant(f, map, inverse_map, tdim):
    """Map H(div) functions."""
    g = subs(f, x, inverse_map)
    J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
    J /= J.det()
    return tuple(vdot(J.row(i), g) for i in range(J.rows))


def double_covariant(f, map, inverse_map, tdim):
    """Map matrix functions."""
    g = subs(f, x, inverse_map)
    J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
    Jinv = J.inv().transpose()

    G = sympy.Matrix([g[i * J.rows: (i + 1) * J.rows] for i in range(J.rows)])
    out = Jinv * G * Jinv.transpose()
    return tuple(out[i] for i in range(out.rows * out.cols))


def double_contravariant(f, map, inverse_map, tdim):
    """Map matrix functions."""
    g = subs(f, x, inverse_map)
    J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
    J /= J.det()

    G = sympy.Matrix([g[i * J.rows: (i + 1) * J.rows] for i in range(J.rows)])
    out = J * G * J.transpose()
    return tuple(out[i] for i in range(out.rows * out.cols))



"""Hellan-Herrmann-Johnson elements on simplices.

This element's definition appears in https://arxiv.org/abs/1909.09687
(Arnold, Walker, 2020)
"""

from ..core.finite_element import FiniteElement
from ..core.moments import make_integral_moment_dofs
from ..core.polynomials import polynomial_set
from ..core.functionals import NormalInnerProductIntegralMoment, IntegralMoment
from .lagrange import DiscontinuousLagrange, SymmetricMatrixDiscontinuousLagrange


class HellanHerrmannJohnson(FiniteElement):
    """A Hellan-Herrmann-Johnson element."""

    def __init__(self, reference, order, variant):
        assert reference.name == "triangle"
        poly = [(p[0], p[1], p[1], p[2])
                for p in polynomial_set(reference.tdim, 3, order)]

        dofs = make_integral_moment_dofs(
            reference,
            facets=(NormalInnerProductIntegralMoment, DiscontinuousLagrange, order),
            cells=(IntegralMoment, SymmetricMatrixDiscontinuousLagrange, order - 1),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))

    names = ["Hellan-Herrmann-Johnson", "HHJ"]
    references = ["triangle"]
    min_order = 0
    mapping = "double_contravariant"
    continuity = "inner H(div)"



"""Serendipity elements on tensor product cells.

This element's definition appears in https://doi.org/10.1007/s10208-011-9087-3
(Arnold, Awanou, 2011)
"""

from ..core.finite_element import FiniteElement
from ..core.moments import make_integral_moment_dofs
from ..core.polynomials import (serendipity_set, polynomial_set,
                                Hdiv_serendipity, Hcurl_serendipity)
from ..core.functionals import (
    PointEvaluation, IntegralMoment, TangentIntegralMoment, NormalIntegralMoment)
from .lagrange import DiscontinuousLagrange, VectorDiscontinuousLagrange


class Serendipity(FiniteElement):
    """A serendipity element."""

    def __init__(self, reference, order, variant):
        poly = polynomial_set(reference.tdim, 1, order)
        poly += serendipity_set(reference.tdim, 1, order)

        dofs = []
        for v_n, p in enumerate(reference.vertices):
            dofs.append(PointEvaluation(p, entity=(0, v_n)))
        dofs += make_integral_moment_dofs(
            reference,
            edges=(IntegralMoment, DiscontinuousLagrange, order - 2),
            faces=(IntegralMoment, DiscontinuousLagrange, order - 4),
            volumes=(IntegralMoment, DiscontinuousLagrange, order - 6),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["serendipity", "S"]
    references = ["interval", "quadrilateral", "hexahedron"]
    min_order = 1
    mapping = "identity"
    continuity = "C0"


class SerendipityCurl(FiniteElement):
    """A serendipity Hcurl element."""

    def __init__(self, reference, order, variant):
        poly = polynomial_set(reference.tdim, reference.tdim, order)
        poly += Hcurl_serendipity(reference.tdim, reference.tdim, order)

        dofs = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DiscontinuousLagrange, order),
            faces=(IntegralMoment, VectorDiscontinuousLagrange, order - 2),
            volumes=(IntegralMoment, VectorDiscontinuousLagrange, order - 4),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["serendipity Hcurl", "Scurl", "BDMCE", "AAE"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    max_order = 3  # TODO: generalise polynomial set for hexahedra, then remove this
    mapping = "covariant"
    continuity = "H(curl)"


class SerendipityDiv(FiniteElement):
    """A serendipity Hdiv element."""

    def __init__(self, reference, order, variant):
        poly = polynomial_set(reference.tdim, reference.tdim, order)
        poly += Hdiv_serendipity(reference.tdim, reference.tdim, order)

        dofs = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DiscontinuousLagrange, order),
            cells=(IntegralMoment, VectorDiscontinuousLagrange, order - 2),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["serendipity Hdiv", "Sdiv", "BDMCF", "AAF"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    max_order = 3  # TODO: generalise polynomial set for hexahedra, then remove this
    mapping = "contravariant"
    continuity = "H(div)"



"""Functions to handle derivatives."""

from .vectors import vdot
from .symbolic import x, sym_sum


def derivative(f, dir):
    """Find the directional derivative of a function."""
    return vdot(grad(f, len(dir)), dir)


def grad(f, dim):
    """Find the gradient of a scalar function."""
    return tuple(f.diff(x[i]) for i in range(dim))


def jacobian_component(f, component):
    """Find a component of the Jacobian."""
    return f.diff(x[component[0]]).diff(x[component[1]])


def jacobian(f, dim):
    """Find the Jacobian."""
    return [[f.diff(x[i]).diff(x[j]) for i in range(dim)] for j in range(dim)]


def div(f):
    """Find the divergence of a vector function."""
    return sym_sum(j.diff(x[i]) for i, j in enumerate(f))


def curl(f):
    """Find the curl of a 3D vector function."""
    return (
        f[2].diff(x[1]) - f[1].diff(x[2]),
        f[0].diff(x[2]) - f[2].diff(x[0]),
        f[1].diff(x[0]) - f[0].diff(x[1])
    )



"""Brezzi-Douglas-Fortin-Marini elements.

This element's definition appears in https://doi.org/10.1051/m2an/1987210405811
(Brezzi, Douglas, Fortin, Marini, 1987)
"""

from ..core.finite_element import FiniteElement
from ..core.moments import make_integral_moment_dofs
from ..core.polynomials import polynomial_set
from ..core.symbolic import x, zero
from ..core.functionals import NormalIntegralMoment, IntegralMoment
from .lagrange import DiscontinuousLagrange, VectorDiscontinuousLagrange


def bdfm_polyset(reference, order):
    """Create the polynomial basis for a BDFM element."""
    dim = reference.tdim
    pset = []
    if reference.name == "quadrilateral":
        for i in polynomial_set(dim, 1, order):
            if i != x[0] ** order:
                pset.append((zero, i))
            if i != x[1] ** order:
                pset.append((i, zero))
    elif reference.name == "triangle":
        pset = polynomial_set(dim, dim, order - 1)
        for i in range(order):
            p = x[0] ** i * x[1] ** (order - 1 - i)
            pset.append((x[0] * p, x[1] * p))
    elif reference.name == "hexahedron":
        pset = polynomial_set(dim, dim, order - 1)
        for i in range(1, order + 1):
            for j in range(order + 1 - i):
                k = order - i - j
                pset.append((x[0] ** i * x[1] ** j * x[2] ** k, zero, zero))
                pset.append((zero, x[1] ** i * x[0] ** j * x[2] ** k, zero))
                pset.append((zero, zero, x[2] ** i * x[0] ** j * x[1] ** k))
    elif reference.name == "tetrahedron":
        pset = polynomial_set(dim, dim, order - 1)
        for i in range(order):
            for j in range(order - i):
                p = x[0] ** i * x[1] ** j * x[2] ** (order - 1 - i - j)
                pset.append((x[0] * p, x[1] * p, x[2] * p))
    return pset


class BDFM(FiniteElement):
    """Brezzi-Douglas-Fortin-Marini Hdiv finite element."""

    def __init__(self, reference, order, variant):
        poly = bdfm_polyset(reference, order)

        dofs = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DiscontinuousLagrange, order - 1),
            cells=(IntegralMoment, VectorDiscontinuousLagrange, order - 2),
            variant=variant
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["Brezzi-Douglas-Fortin-Marini", "BDFM"]
    references = ["triangle", "quadrilateral", "hexahedron", "tetrahedron"]
    min_order = 1
    mapping = "contravariant"
    continuity = "H(div)"



"""Quadrature definitions."""

import sympy
from ..core.symbolic import one, zero


def equispaced(N):
    """Get equispaced points and weights.

    Parameters
    ----------
    N : int
        Number of points
    """
    return ([sympy.Rational(i, N - 1) for i in range(N)],
            [sympy.Rational(1, 2*(N-1)) if i == 0 or i == N - 1 else sympy.Rational(1, N-1)
             for i in range(N)])


def lobatto(N):
    """Get Gauss-Lobatto-Legendre points and weights.

    Parameters
    ----------
    N : int
        Number of points
    """
    if N == 2:
        return ([zero, one],
                [sympy.Rational(1, 2), sympy.Rational(1, 2)])
    if N == 3:
        return ([zero, sympy.Rational(1, 2), one],
                [sympy.Rational(1, 6), sympy.Rational(2, 3), sympy.Rational(1, 6)])
    if N == 4:
        return ([zero, (1 - 1 / sympy.sqrt(5)) / 2, (1 + 1 / sympy.sqrt(5)) / 2, one],
                [sympy.Rational(1, 12), sympy.Rational(5, 12), sympy.Rational(5, 12),
                 sympy.Rational(1, 12)])
    if N == 5:
        return ([zero, (1 - sympy.sqrt(3) / sympy.sqrt(7)) / 2, sympy.Rational(1, 2),
                 (1 + sympy.sqrt(3) / sympy.sqrt(7)) / 2, one],
                [sympy.Rational(1, 20), sympy.Rational(49, 180), sympy.Rational(16, 45),
                 sympy.Rational(49, 180), sympy.Rational(1, 20)])
    if N == 6:
        return ([zero,
                 (1 - sympy.sqrt(sympy.Rational(1, 3) + (2 * sympy.sqrt(7) / 21))) / 2,
                 (1 - sympy.sqrt(sympy.Rational(1, 3) - (2 * sympy.sqrt(7) / 21))) / 2,
                 (1 + sympy.sqrt(sympy.Rational(1, 3) - (2 * sympy.sqrt(7) / 21))) / 2,
                 (1 + sympy.sqrt(sympy.Rational(1, 3) + (2 * sympy.sqrt(7) / 21))) / 2,
                 one],
                [sympy.Rational(1, 30), (14 - sympy.sqrt(7)) / 60, (14 + sympy.sqrt(7)) / 60,
                 (14 + sympy.sqrt(7)) / 60, (14 - sympy.sqrt(7)) / 60, sympy.Rational(1, 30)])
    if N == 7:
        return ([zero,
                 (1 - sympy.sqrt((5 + 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                 (1 - sympy.sqrt((5 - 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                 sympy.Rational(1, 2),
                 (1 + sympy.sqrt((5 - 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                 (1 + sympy.sqrt((5 + 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                 one],
                [sympy.Rational(1, 42),
                 (124 - 7 * sympy.sqrt(15)) / 700,
                 (124 + 7 * sympy.sqrt(15)) / 700,
                 sympy.Rational(128, 525),
                 (124 + 7 * sympy.sqrt(15)) / 700,
                 (124 - 7 * sympy.sqrt(15)) / 700,
                 sympy.Rational(1, 42)])
    raise NotImplementedError()


def radau(N):
    """Get Radau points and weights.

    Parameters
    ----------
    N : int
        Number of points
    """
    if N == 2:
        return ([zero, sympy.Rational(2, 3)],
                [sympy.Rational(1, 4), sympy.Rational(3, 4)])
    if N == 3:
        return ([zero, (6 - sympy.sqrt(6)) / 10, (6 + sympy.sqrt(6)) / 10],
                [sympy.Rational(1, 9), (16 + sympy.sqrt(6)) / 36, (16 - sympy.sqrt(6)) / 36])
    raise NotImplementedError()


def legendre(N):
    """Get Gauss-Legendre points and weights.

    Parameters
    ----------
    N : int
        Number of points
    """
    if N == 1:
        return ([sympy.Rational(1, 2)], [one])
    if N == 2:
        return ([(3 - sympy.sqrt(3)) / 6, (3 + sympy.sqrt(3)) / 6],
                [sympy.Rational(1, 2), sympy.Rational(1, 2)])
    if N == 3:
        return ([(5 - sympy.sqrt(15)) / 10, sympy.Rational(1, 2), (5 + sympy.sqrt(15)) / 10],
                [sympy.Rational(5, 18), sympy.Rational(4, 9), sympy.Rational(5, 18)])
    raise NotImplementedError()


def get_quadrature(rule, N):
    """Get quadrature points and weights.

    Parameters
    ----------
    rule : str
        The quadrature rule.
        Supported values: equispaced, lobatto, radau, legendre
    N : int
        Number of points
    """
    if rule == "equispaced":
        return equispaced(N)
    if rule == "lobatto":
        return lobatto(N)
    if rule == "radau":
        return radau(N)
    if rule == "legendre":
        return legendre(N)
    raise ValueError(f"Unknown quadrature rule: {rule}")



"""Functions to handle vectors."""

import sympy


def vsub(v, w):
    """Subtract a vector from another."""
    try:
        return tuple(i - j for i, j in zip(v, w))
    except TypeError:
        return v - w


def vadd(v, w):
    """Add two vectors."""
    try:
        return tuple(i + j for i, j in zip(v, w))
    except TypeError:
        return v + w


def vdiv(v, a):
    """Divide a vector by a scalar."""
    try:
        return tuple(i / a for i in v)
    except TypeError:
        return v / a


def vnorm(v):
    """Find the norm of a vector."""
    try:
        return sympy.sqrt(sum(a ** 2 for a in v))
    except TypeError:
        return abs(v)


def vdot(v, w):
    """Find the dot product of two vectors."""
    try:
        return sum(a * b for a, b in zip(v, w))
    except TypeError:
        return v * w


def vcross(v, w):
    """Find the cross product of two vectors."""
    if len(v) == 2:
        return _vcross2d(v, w)
    else:
        assert len(v) == 3
        return _vcross3d(v, w)


def _vcross2d(v, w):
    """Find the cross product of two 2D vectors."""
    return v[0] * w[1] - v[1] * w[0]


def _vcross3d(v, w):
    """Find the cross product of two 3D vectors."""
    return (
        v[1] * w[2] - v[2] * w[1],
        v[2] * w[0] - v[0] * w[2],
        v[0] * w[1] - v[1] * w[0],
    )


def vnormalise(v):
    """Normalise a vector."""
    return vdiv(v, vnorm(v))


def point_in_triangle(point, triangle):
    """Check if a point is inside a triangle."""
    v0 = vsub(triangle[2], triangle[0])
    v1 = vsub(triangle[1], triangle[0])
    v2 = vsub(point, triangle[0])

    dot00 = vdot(v0, v0)
    dot01 = vdot(v0, v1)
    dot02 = vdot(v0, v2)
    dot11 = vdot(v1, v1)
    dot12 = vdot(v1, v2)

    det = (dot00 * dot11 - dot01 * dot01)
    u = (dot11 * dot02 - dot01 * dot12) / det
    v = (dot00 * dot12 - dot01 * dot02) / det

    return u >= 0 and v >= 0 and u + v < 1



"""Symfem: a symbolic finite element definition library."""
import os as _os
import importlib as _il
from .core import references as _references
from .core.finite_element import FiniteElement as _FiniteElement

_folder = _os.path.dirname(_os.path.realpath(__file__))

if _os.path.isfile(_os.path.join(_folder, "../VERSION")):
    # If running from folder
    _v_folder = _os.path.join(_folder, "..")
else:
    # If running from installation
    _v_folder = _folder

with open(_os.path.join(_v_folder, "VERSION")) as _f:
    __version__ = _f.read().strip()

_elementmap = {}
_elementlist = []

for _file in _os.listdir(_os.path.join(_folder, "elements")):
    if _file.endswith(".py") and "__init__" not in _file:
        _fname = _file[:-3]
        _module = _il.import_module(f"symfem.elements.{_fname}")

        for _class_name in dir(_module):
            _element = getattr(_module, _class_name)
            if (
                isinstance(_element, type)
                and issubclass(_element, _FiniteElement)
                and _element != _FiniteElement
                and len(_element.names) > 0
            ):
                if _element not in _elementlist:
                    _elementlist.append(_element)
                for _n in _element.names:
                    if _n in _elementmap:
                        assert _element == _elementmap[_n]
                    _elementmap[_n] = _element


def create_reference(cell_type, vertices=None):
    """Make a reference cell.

    Parameters
    ----------
    cell_type : str
        The reference cell type.
        Supported values: interval, triangle, quadrilateral, tetrahedron, hexahedron,
        dual polygon(number_of_triangles)
    vertices : list
        The vertices of the reference.
    """
    if cell_type == "interval":
        if vertices is not None:
            return _references.Interval(vertices)
        return _references.Interval()
    elif cell_type == "triangle":
        if vertices is not None:
            return _references.Triangle(vertices)
        return _references.Triangle()
    elif cell_type == "tetrahedron":
        if vertices is not None:
            return _references.Tetrahedron(vertices)
        return _references.Tetrahedron()
    elif cell_type == "quadrilateral":
        if vertices is not None:
            return _references.Quadrilateral(vertices)
        return _references.Quadrilateral()
    elif cell_type == "hexahedron":
        if vertices is not None:
            return _references.Hexahedron(vertices)
        return _references.Hexahedron()
    elif cell_type.startswith("dual polygon"):
        n_tri = int(cell_type.split("(")[1].split(")")[0])
        if vertices is not None:
            return _references.DualPolygon(n_tri, vertices)
        return _references.DualPolygon(n_tri)
    else:
        raise ValueError(f"Unknown cell type: {cell_type}")


def create_element(cell_type, element_type, order, variant="equispaced"):
    """Make a finite element.

    Parameters
    ----------
    cell_type : str
        The reference cell type.
        Supported values: interval, triangle, quadrilateral, tetrahedron, hexahedron,
        dual polygon(number_of_triangles)
    element_type : str
        The type of the element.
        Supported values: Lagrange, P, discontinuous Lagrange, dP, DP,
        vector Lagrange, vP, vector discontinuous Lagrange, vdP, vDP,
        matrix discontinuous Lagrange, symmetric matrix discontinuous Lagrange,
        Crouzeix-Raviart, CR, serendipity, S, serendipity Hcurl, Scurl, BDMCE, AAE,
        serendipity Hdiv, Sdiv, BDMCF, AAF,  Regge, Nedelec, Nedelec1, N1curl, Nedelec2,
        N2curl, Raviart-Thomas, RT, N1div, dQ, NCE, RTCE, Qcurl, Q, NCF, RTCF, Qdiv,
        vector Q, vQ, Brezzi-Douglas-Marini, BDM, N2div, Morley, Hermite,
        Mardal-Tai-Winther, MTW, Argyris, bubble, dual, Buffa-Christiansen, BC,
        rotated Buffa-Christiansen, RBC, Brezzi-Douglas-Fortin-Marini, BDFM,
        Brezzi-Douglas-Duran-Fortin, BDDF, Hellan-Herrmann-Johnson, HHJ,
        Arnold-Winther, AW, Bell, Kong-Mulder-Veldhuizen, KMV,
        Bernstein, Bernstein-Bezier,
        Hsieh-Clough-Tocher, Clough-Tocher, HCT, CT,
        reduced Hsieh-Clough-Tocher, rHCT
    order : int
        The order of the element.
    variant : str
        The arrangement type of the points used the define the space.
        Supported values: equispaced, lobatto, radau, legendre
    """
    reference = create_reference(cell_type)

    if element_type in _elementmap:
        assert cell_type.split("(")[0] in _elementmap[element_type].references
        return _elementmap[element_type](reference, order, variant=variant)

    raise ValueError(f"Unsupported element type: {element_type}")



import symfem
from io import StringIO
import sys
import os
import re
import pytest

code = False
output = False
lines = []
doc_data = []
outputlines = []
codelines = []

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../docs/index.rst")) as f:
    for line in f:
        if line.strip() == ".. code-block:: python":
            code = True
            output = False
            codelines = []
            outputlines = []
        elif code and line.strip() == "::":
            output = True
            code = False
        elif code or output:
            if line.strip() != "":
                if line.startswith("    "):
                    if code:
                        assert not output
                        codelines.append(line.strip("\n")[4:])
                    else:
                        assert output
                        outputlines.append(line.strip("\n")[4:])
                else:
                    output = False
                    code = False
                    doc_data.append(("\n".join(codelines), "\n".join(outputlines)))


def test_available_references():
    r_supported = symfem.create_reference.__doc__
    r_supported = r_supported.split("cell_type : str\n")[1]
    r_supported = r_supported.split("vertices : list\n")[0]
    r_supported = r_supported.split("Supported values:")[1]
    r_supported = set([i.strip() for i in r_supported.split(",")])

    e_supported = symfem.create_element.__doc__
    e_supported = e_supported.split("cell_type : str\n")[1]
    e_supported = e_supported.split("element_type : str\n")[0]
    e_supported = e_supported.split("Supported values:")[1]
    e_supported = set([i.strip() for i in e_supported.split(",")])

    assert r_supported == e_supported


def test_available_variants():
    r_supported = symfem.core.quadrature.get_quadrature.__doc__
    r_supported = r_supported.split("rule : str\n")[1]
    r_supported = r_supported.split("N : int\n")[0]
    r_supported = r_supported.split("Supported values:")[1]
    r_supported = set([i.strip() for i in r_supported.split(",")])

    e_supported = symfem.create_element.__doc__
    e_supported = e_supported.split("variant : str\n")[1]
    e_supported = e_supported.split("Supported values:")[1]
    e_supported = set([i.strip() for i in e_supported.split(",")])

    assert r_supported == e_supported


def test_available_elements():
    supported = symfem.create_element.__doc__
    supported = supported.split("element_type : str\n")[1]
    supported = supported.split("order : int\n")[0]
    supported = supported.split("Supported values:")[1]
    supported = set([i.strip() for i in supported.split(",")])

    assert set(symfem._elementmap.keys()) == supported


@pytest.mark.parametrize("script, output", doc_data)
def test_snippets(script, output):
    old_stdout = sys.stdout
    redirected_output = sys.stdout = StringIO()
    exec(script)
    sys.stdout = old_stdout

    actual_output = redirected_output.getvalue().strip()
    actual_output = re.sub(r"at 0x[^\>]+\>", "at 0x{ADDRESS}>", actual_output)
    assert actual_output == output



"""Kong-Mulder-Veldhuizen elements on triangle.

This element's definition is given in https://doi.org/10.1023/A:1004420829610
(Chin-Joe-Kong, Mulder, Van Veldhuizen, 1999)
"""

import sympy
from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import WeightedPointEvaluation
from ..core.symbolic import x


def kmv_tri_polyset(m, mf):
    """Create the polynomial set for a KMV space on a triangle."""
    poly = polynomial_set(2, 1, m)

    b = x[0] * x[1] * (1 - x[0] - x[1])
    poly += [x[0] ** p * x[1] ** (mf - 3 - p) * b
             for p in range(mf - 2)]

    return poly


def kmv_tet_polyset(m, mf, mi):
    """Create the polynomial set for a KMV space on a tetrahedron."""
    poly = polynomial_set(3, 1, m)

    # TODO: check this
    for axes in [(x[0], x[1]), (x[0], x[2]), (x[1], x[2]), (x[1] - x[0], x[2] - x[0])]:
        b = axes[0] * axes[1] * (1 - axes[0] - axes[1])
        for i in range(mf - 2):
            for j in range(mf - 2 - i):
                poly.append(x[0] ** i * x[1] ** j * x[2] ** (mf - 3 - i - j) * b)

    b = x[0] * x[1] * x[2] * (1 - x[0] - x[1] - x[2])
    for i in range(mi - 3):
        for j in range(mi - 3 - i):
            poly.append(x[0] ** i * x[1] ** j * x[2] ** (mf - 4 - i - j) * b)

    return poly


class KongMulderVeldhuizen(FiniteElement):
    """Kong-Mulder-Veldhuizen finite element."""

    def __init__(self, reference, order, variant):
        dofs = []
        if reference.name == "triangle":
            if order == 1:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(WeightedPointEvaluation(v, sympy.Rational(1, 6), entity=(0, v_n)))
                poly = kmv_tri_polyset(1, 1)

            elif order == 3:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(WeightedPointEvaluation(v, sympy.Rational(1, 40), entity=(0, v_n)))
                for e_n, e in enumerate(reference.edges):
                    midpoint = tuple(sympy.Rational(i + j,  2) for i, j in zip(
                        reference.vertices[e[0]], reference.vertices[e[1]]))
                    dofs.append(WeightedPointEvaluation(
                        midpoint, sympy.Rational(1, 15), entity=(1, e_n)))
                dofs.append(WeightedPointEvaluation(
                    (sympy.Rational(1, 3), sympy.Rational(1, 3)), sympy.Rational(9, 40),
                    entity=(2, 0)))
                poly = kmv_tri_polyset(2, 3)

            elif order == 4:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(WeightedPointEvaluation(
                        v, sympy.Rational(1, 90) - sympy.sqrt(7) / 720, entity=(0, v_n)))
                alpha = sympy.Rational(1, 2) - sympy.sqrt(411 - 84 * (7 - sympy.sqrt(7))) / 42
                for e_n, e in enumerate(reference.edges):
                    dofs.append(WeightedPointEvaluation(
                        tuple(i + (j - i) * alpha for i, j in zip(
                            reference.vertices[e[0]], reference.vertices[e[1]])),
                        sympy.Rational(7, 720) - sympy.sqrt(7) / 180, entity=(1, e_n)))
                    dofs.append(WeightedPointEvaluation(
                        tuple(j + (i - j) * alpha for i, j in zip(
                            reference.vertices[e[0]], reference.vertices[e[1]])),
                        sympy.Rational(7, 720) - sympy.sqrt(7) / 180, entity=(1, e_n)))
                beta = (1 - 1 / sympy.sqrt(7)) / 3
                for i in [(0, 1, 2), (1, 2, 0), (2, 0, 1)]:
                    dofs.append(WeightedPointEvaluation(
                        tuple(a + beta * (b - a) + beta * (c - a)
                              for a, b, c in zip(*[reference.vertices[j] for j in i])),
                        sympy.Rational(49, 360) - 7 * sympy.sqrt(7) / 720,
                        entity=(2, 0)))
                poly = kmv_tri_polyset(3, 4)

            else:
                raise NotImplementedError

        elif reference.name == "tetrahedron":
            if order == 1:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(WeightedPointEvaluation(v, sympy.Rational(1, 24), entity=(0, v_n)))
                poly = kmv_tet_polyset(1, 1, 1)

            elif order == 4:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(WeightedPointEvaluation(v, (13 - 3 * sympy.sqrt(13)) / 10080,
                                                        entity=(0, v_n)))
                for e_n, e in enumerate(reference.edges):
                    midpoint = tuple(sympy.Rational(i + j,  2) for i, j in zip(
                        reference.vertices[e[0]], reference.vertices[e[1]]))
                    dofs.append(WeightedPointEvaluation(
                        midpoint, (4 - sympy.sqrt(13)) / 315, entity=(1, e_n)))
                alpha = (7 - sympy.sqrt(13)) / 18
                for f_n, face in enumerate(reference.faces):
                    for i in [(0, 1, 2), (1, 2, 0), (2, 0, 1)]:
                        dofs.append(WeightedPointEvaluation(
                            tuple(a + alpha * (b - a) + alpha * (c - a)
                                  for a, b, c in zip(*[reference.vertices[face[j]] for j in i])),
                            (29 + 17 * sympy.sqrt(13)) / 10080,
                            entity=(2, f_n)))

                dofs.append(WeightedPointEvaluation(
                    tuple(sympy.Rational(1, 4) for i in range(3)), sympy.Rational(16, 315),
                    entity=(3, 0)))
                poly = kmv_tet_polyset(2, 4, 4)

            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["Kong-Mulder-Veldhuizen", "KMV"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "C0"
    mapping = "identity"



"""Hermite elements on simplices.

This element's definition appears in https://doi.org/10.1016/0045-7825(72)90006-0
(Ciarlet, Raviart, 1972)
"""

import sympy
from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import PointEvaluation, DerivativePointEvaluation
from ..core.symbolic import sym_sum, x, subs


class Hermite(FiniteElement):
    """Hermite finite element."""

    def __init__(self, reference, order, variant):
        assert order == 3
        dofs = []
        for v_n, vs in enumerate(reference.sub_entities(0)):
            dofs.append(PointEvaluation(vs, entity=(0, v_n)))
            for i in range(reference.tdim):
                dofs.append(DerivativePointEvaluation(
                    vs, tuple(1 if i == j else 0 for j in range(reference.tdim)),
                    entity=(0, v_n)))
        for e_n, vs in enumerate(reference.sub_entities(2)):
            midpoint = tuple(sym_sum(i) / len(i)
                             for i in zip(*[reference.vertices[i] for i in vs]))
            dofs.append(PointEvaluation(midpoint, entity=(2, e_n)))

        super().__init__(
            reference, order, polynomial_set(reference.tdim, 1, order), dofs, reference.tdim, 1
        )

    def perform_mapping(self, basis, map, inverse_map):
        """Map the basis onto a cell using the appropriate mapping for the element."""
        out = []
        tdim = self.reference.tdim
        J = sympy.Matrix([[map[i].diff(x[j]) for j in range(tdim)] for i in range(tdim)])
        for v in range(tdim + 1):
            out.append(basis[(tdim + 1) * v])
            for i in range(tdim):
                out.append(sym_sum(a * b for a, b in
                                   zip(basis[(tdim + 1) * v + 1: (tdim + 1) * (v + 1)],
                                       J.row(i))))
        if tdim == 2:
            out.append(basis[-1])
        if tdim == 3:
            out += basis[-4:]
        assert len(out) == len(basis)
        return [subs(b, x, inverse_map) for b in out]

    names = ["Hermite"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 3
    max_order = 3
    continuity = "C0"



from symfem import create_element
from symfem.core.symbolic import subs, x
import pytest

elements = {
    "interval": [("P", "Lagrange", range(1, 4)), ("dP", "Discontinuous Lagrange", range(1, 4))],
    "triangle": [("P", "Lagrange", range(1, 4)), ("dP", "Discontinuous Lagrange", range(1, 4)),
                 ("N1curl", "Nedelec 1st kind H(curl)", range(1, 4)),
                 ("N2curl", "Nedelec 2nd kind H(curl)", range(1, 4)),
                 ("N1div", "Raviart-Thomas", range(1, 4)),
                 ("N2div", "Brezzi-Douglas-Marini", range(1, 4)),
                 ("Regge", "Regge", range(0, 4)),
                 ("Crouzeix-Raviart", "Crouzeix-Raviart", [1])],
    "tetrahedron": [("P", "Lagrange", range(1, 4)), ("dP", "Discontinuous Lagrange", range(1, 4)),
                    ("N1curl", "Nedelec 1st kind H(curl)", range(1, 4)),
                    ("N2curl", "Nedelec 2nd kind H(curl)", range(1, 3)),
                    ("N1div", "Raviart-Thomas", range(1, 3)),
                    ("N2div", "Brezzi-Douglas-Marini", range(1, 3)),
                    ("Regge", "Regge", range(0, 3)),
                    ("Crouzeix-Raviart", "Crouzeix-Raviart", [1])],
    "quadrilateral": [("Q", "Lagrange", range(1, 4)),
                      ("dQ", "Discontinuous Lagrange", range(1, 4)),
                      ("Qdiv", "Raviart-Thomas", range(1, 4)),
                      ("Qcurl", "Nedelec 1st kind H(curl)", range(1, 4))],
    "hexahedron": [("Q", "Lagrange", range(1, 3)), ("dQ", "Discontinuous Lagrange", range(1, 3)),
                   ("Qdiv", "Raviart-Thomas", range(1, 3)),
                   ("Qcurl", "Nedelec 1st kind H(curl)", range(1, 3))]
}


def to_float(a):
    try:
        return float(a)
    except:  # noqa: E722
        return [to_float(i) for i in a]


def make_lattice(cell, N=3):
    import numpy as np
    if cell == "interval":
        return np.array([[i / N] for i in range(N + 1)])
    if cell == "triangle":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1 - i)])
    if cell == "tetrahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i)
                         for k in range(N + 1 - i - j)])
    if cell == "quadrilateral":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1)])
    if cell == "hexahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1) for k in range(N + 1)])


@pytest.mark.parametrize(("cell", "symfem_type", "basix_type", "order"),
                         [(cell, a, b, order) for cell, ls in elements.items()
                          for a, b, c in ls for order in c])
def test_against_basix(cell, symfem_type, basix_type, order):
    try:
        import basix
        from scipy.linalg import block_diag, solve
        import numpy as np
    except ImportError:
        pytest.skip("basix, numpy and scipy must be installed to run this test.")

    element = create_element(cell, symfem_type, order)
    points = make_lattice(cell)
    space = basix.create_element(basix_type, cell, order)
    result = space.tabulate(0, points)[0]

    mat = element.get_dual_matrix()
    mat = np.array([[float(j) for j in mat.row(i)] for i in range(mat.rows)])

    if element.range_dim == 1:
        evaluated = np.array([[float(subs(b, x, p)) for p in points] for b in element.basis])
    else:
        evaluated = np.array([[float(subs(b, x, p)[j]) for p in points]
                              for j in range(element.range_dim) for b in element.basis])
        mat = block_diag(*[mat for j in range(element.range_dim)])

    sym_result = solve(mat, evaluated).transpose()
    assert np.allclose(result, sym_result)



"""Morley elements on simplices.

This element's definition appears in https://doi.org/10.1017/S0001925900004546
(Morley, 1968)
"""

from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import PointEvaluation, PointNormalDerivativeEvaluation
from ..core.symbolic import sym_sum


class Morley(FiniteElement):
    """Morley finite element."""

    def __init__(self, reference, order, variant):
        from symfem import create_reference
        assert order == 2
        assert reference.name == "triangle"
        dofs = []
        for v_n, vs in enumerate(reference.sub_entities(0)):
            dofs.append(PointEvaluation(vs, entity=(0, v_n)))
        for e_n, vs in enumerate(reference.sub_entities(1)):
            sub_ref = create_reference(
                reference.sub_entity_types[1],
                vertices=[reference.reference_vertices[v] for v in vs])
            midpoint = tuple(sym_sum(i) / len(i)
                             for i in zip(*[reference.vertices[i] for i in vs]))
            dofs.append(
                PointNormalDerivativeEvaluation(midpoint, sub_ref, entity=(1, e_n)))

        super().__init__(
            reference, order, polynomial_set(reference.tdim, 1, order), dofs, reference.tdim, 1
        )

    names = ["Morley"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    mapping = "identity"
    continuity = "L2"



"""Bubble elements on simplices.

This element's definition appears in https://doi.org/10.1007/978-3-642-23099-8_3
(Kirby, Logg, Rognes, Terrel, 2012)
"""

import sympy
from itertools import product
from ..core.symbolic import x
from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import PointEvaluation


class Bubble(FiniteElement):
    """Bubble finite element."""

    def __init__(self, reference, order, variant):
        if reference.name == "interval":
            poly = [x[0] * (1 - x[0]) * p for p in polynomial_set(reference.tdim, 1, order - 2)]
        elif reference.name == "triangle":
            poly = [x[0] * x[1] * (1 - x[0] - x[1]) * p
                    for p in polynomial_set(reference.tdim, 1, order - 3)]
        else:
            assert reference.name == "tetrahedron"
            poly = [x[0] * x[1] * x[2] * (1 - x[0] - x[1] - x[2]) * p
                    for p in polynomial_set(reference.tdim, 1, order - 4)]

        dofs = []
        for i in product(range(1, order), repeat=reference.tdim):
            if sum(i) < order:
                dofs.append(
                    PointEvaluation(
                        tuple(o + sum(sympy.Rational(a[j] * b, order)
                                      for a, b in zip(reference.axes, i))
                              for j, o in enumerate(reference.origin)),
                        entity=(reference.tdim, 0)))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["bubble"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = {"interval": 2, "triangle": 3, "tetrahedron": 4}
    mapping = "identity"
    continuity = "C0"



"""Functionals used to define the dual sets."""
from .symbolic import subs, x, t, PiecewiseFunction
from .vectors import vdot
from .calculus import derivative, jacobian_component


class BaseFunctional:
    """A functional."""

    def __init__(self, entity=(None, None)):
        self.entity = entity

    def eval(self, fun):
        """Apply to the functional to a function."""
        raise NotImplementedError

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return tuple(None for i in range(self.reference.gdim))

    def dof_direction(self):
        """Get the direction of the DOF."""
        return None

    def entity_dim(self):
        """Get the dimension of the entitiy this DOF is associated with."""
        return self.entity[0]

    name = None


class PointEvaluation(BaseFunctional):
    """A point evaluation."""

    def __init__(self, point, entity=(None, None)):
        super().__init__(entity)
        self.point = point

    def eval(self, function):
        """Apply to the functional to a function."""
        return subs(function, x, self.point)

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    name = "Point evaluation"


class WeightedPointEvaluation(BaseFunctional):
    """A point evaluation."""

    def __init__(self, point, weight, entity=(None, None)):
        super().__init__(entity)
        self.point = point
        self.weight = weight

    def eval(self, function):
        """Apply to the functional to a function."""
        return subs(function, x, self.point) * self.weight

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    name = "Weighted point evaluation"


class DerivativePointEvaluation(BaseFunctional):
    """A point evaluation of a given derivative."""

    def __init__(self, point, derivative, entity=(None, None)):
        super().__init__(entity)
        self.point = point
        self.derivative = derivative

    def eval(self, function):
        """Apply to the functional to a function."""
        for i, j in zip(x, self.derivative):
            for k in range(j):
                function = function.diff(i)
        return subs(function, x, self.point)

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    name = "Point derivative evaluation"


class PointDirectionalDerivativeEvaluation(BaseFunctional):
    """A point evaluation of a derivative in a fixed direction."""

    def __init__(self, point, direction, entity=(None, None)):
        super().__init__(entity)
        self.point = point
        self.dir = direction

    def eval(self, function):
        """Apply to the functional to a function."""
        if isinstance(function, PiecewiseFunction):
            function = function.get_piece(self.point)
        return subs(derivative(function, self.dir), x, self.point)

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    def dof_direction(self):
        """Get the direction of the DOF."""
        return self.dir

    name = "Point evaluation of directional derivative"


class PointNormalDerivativeEvaluation(PointDirectionalDerivativeEvaluation):
    """A point evaluation of a normal derivative."""

    def __init__(self, point, edge, entity=(None, None)):
        super().__init__(point, edge.normal(), entity=entity)
        self.reference = edge

    name = "Point evaluation of normal derivative"


class PointComponentSecondDerivativeEvaluation(BaseFunctional):
    """A point evaluation of a component of a second derivative."""

    def __init__(self, point, component, entity=(None, None)):
        super().__init__(entity)
        self.point = point
        self.component = component

    def eval(self, function):
        """Apply to the functional to a function."""
        return subs(jacobian_component(function, self.component), x, self.point)

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    name = "Point evaluation of Jacobian component"


class PointInnerProduct(BaseFunctional):
    """An evaluation of an inner product at a point."""

    def __init__(self, point, lvec, rvec, entity=(None, None)):
        super().__init__(entity)
        self.point = point
        self.lvec = lvec
        self.rvec = rvec

    def eval(self, function):
        """Apply to the functional to a function."""
        v = subs(function, x, self.point)
        tdim = len(self.lvec)
        assert len(function) == tdim ** 2
        return vdot(self.lvec,
                    tuple(vdot(v[tdim * i: tdim * (i + 1)], self.rvec)
                          for i in range(0, tdim)))

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    def dof_direction(self):
        """Get the location of the DOF in the cell."""
        if self.rvec != self.lvec:
            return None
        return self.lvec

    name = "Point inner product"


class DotPointEvaluation(BaseFunctional):
    """A point evaluation in a given direction."""

    def __init__(self, point, vector, entity=(None, None)):
        super().__init__(entity)
        self.point = point
        self.vector = vector

    def eval(self, function):
        """Apply to the functional to a function."""
        return subs(vdot(function, self.vector), x, self.point)

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        return self.point

    def dof_direction(self):
        """Get the direction of the DOF."""
        return self.vector

    name = "Dot point evaluation"


class IntegralAgainst(BaseFunctional):
    """An integral against a function."""

    def __init__(self, reference, f, entity=(None, None)):
        super().__init__(entity)
        self.reference = reference
        self.f = subs(f, x, t)
        if isinstance(self.f, tuple):
            if len(self.f) == self.reference.tdim:
                # TODO: is this one of the mappings?
                self.f = tuple(
                    sum(self.reference.axes[j][i] * c / self.reference.jacobian()
                        for j, c in enumerate(self.f))
                    for i, o in enumerate(self.reference.origin)
                )
            else:
                assert len(self.f) == self.reference.tdim ** 2
                assert self.reference.vertices == self.reference.reference_vertices

    def eval(self, function):
        """Apply to the functional to a function."""
        point = [i for i in self.reference.origin]
        for i, a in enumerate(zip(*self.reference.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        integrand = self.dot(subs(function, x, point))
        return self.reference.integral(integrand)

    def dot(self, function):
        """Dot a function with the moment function."""
        return vdot(function, self.f)

    name = "Integral against"


class IntegralMoment(BaseFunctional):
    """An integral moment."""

    def __init__(self, reference, f, dof, entity=(None, None)):
        super().__init__(entity)
        self.reference = reference
        self.dof = dof
        self.f = subs(f, x, t)
        if isinstance(self.f, tuple):
            if len(self.f) == self.reference.tdim:
                # TODO: is this one of the mappings?
                self.f = tuple(
                    sum(self.reference.axes[j][i] * c / self.reference.jacobian()
                        for j, c in enumerate(self.f))
                    for i, o in enumerate(self.reference.origin)
                )
            else:
                assert len(self.f) == self.reference.tdim ** 2
                assert self.reference.vertices == self.reference.reference_vertices

    def eval(self, function):
        """Apply to the functional to a function."""
        point = [i for i in self.reference.origin]
        for i, a in enumerate(zip(*self.reference.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        integrand = self.dot(subs(function, x, point))
        return self.reference.integral(integrand)

    def dot(self, function):
        """Dot a function with the moment function."""
        return vdot(function, self.f)

    def dof_point(self):
        """Get the location of the DOF in the cell."""
        p = self.dof.dof_point()
        return tuple(
            o + sum(self.reference.axes[j][i] * c for j, c in enumerate(p))
            for i, o in enumerate(self.reference.origin)
        )

    def dof_direction(self):
        """Get the direction of the DOF."""
        p = self.dof.dof_direction()
        if p is None:
            return None
        return tuple(
            sum(self.reference.axes[j][i] * c for j, c in enumerate(p))
            for i in range(self.reference.gdim)
        )

    name = "Integral moment"


class VecIntegralMoment(IntegralMoment):
    """An integral moment applied to a component of a vector."""

    def __init__(self, reference, f, dot_with, dof, entity=(None, None)):
        super().__init__(reference, f, dof, entity=entity)
        self.dot_with = dot_with

    def dot(self, function):
        """Dot a function with the moment function."""
        return vdot(function, self.dot_with) * self.f

    def dof_direction(self):
        """Get the direction of the DOF."""
        return self.dot_with

    name = "Vector integral moment"


class TangentIntegralMoment(VecIntegralMoment):
    """An integral moment in the tangential direction."""

    def __init__(self, reference, f, dof, entity=(None, None)):
        super().__init__(reference, f, reference.tangent(), dof, entity=entity)

    name = "Tangential integral moment"


class NormalIntegralMoment(VecIntegralMoment):
    """An integral moment in the normal direction."""

    def __init__(self, reference, f, dof, entity=(None, None)):
        super().__init__(reference, f, reference.normal(), dof, entity=entity)

    name = "Normal integral moment"


class InnerProductIntegralMoment(IntegralMoment):
    """An integral moment of the inner product with a vector."""

    def __init__(self, reference, f, inner_with_left, inner_with_right, dof, entity=(None, None)):
        super().__init__(reference, f, dof, entity=entity)
        self.inner_with_left = inner_with_left
        self.inner_with_right = inner_with_right

    def dot(self, function):
        """Take the inner product of a function with the moment direction."""
        tdim = len(self.inner_with_left)
        return vdot(self.inner_with_left,
                    tuple(vdot(function[tdim * i: tdim * (i + 1)], self.inner_with_right)
                          for i in range(0, tdim))) * self.f * self.reference.jacobian()

    def dof_direction(self):
        """Get the direction of the DOF."""
        if self.inner_with_left != self.inner_with_right:
            return None
        return self.inner_with_left

    name = "Inner product integral moment"


class NormalInnerProductIntegralMoment(InnerProductIntegralMoment):
    """An integral moment of the inner product with the normal direction."""

    def __init__(self, reference, f, dof, entity=(None, None)):
        super().__init__(reference, f, reference.normal(), reference.normal(), dof, entity=entity)

    name = "Normal inner product integral moment"



"""Crouzeix-Raviart elements on simplices.

This element's definition appears in https://doi.org/10.1051/m2an/197307R300331
(Crouzeix, Raviart, 1973)
"""

from ..core.finite_element import FiniteElement
from ..core.polynomials import polynomial_set
from ..core.functionals import PointEvaluation
from ..core.symbolic import sym_sum


class CrouzeixRaviart(FiniteElement):
    """Crouzeix-Raviart finite element."""

    def __init__(self, reference, order, variant):
        assert order == 1
        dofs = []
        for e_n, vs in enumerate(reference.sub_entities(reference.tdim - 1)):
            midpoint = tuple(sym_sum(i) / len(i)
                             for i in zip(*[reference.vertices[i] for i in vs]))
            dofs.append(
                PointEvaluation(midpoint, entity=(reference.tdim - 1, e_n)))

        super().__init__(
            reference, order, polynomial_set(reference.tdim, 1, order), dofs, reference.tdim, 1
        )

    names = ["Crouzeix-Raviart", "CR"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = 1
    mapping = "identity"
    continuity = "L2"



import sys
import os
import setuptools

os.system("cp VERSION symfem/")

if sys.version_info < (3, 6):
    print("Python 3.6 or higher required, please upgrade.")
    sys.exit(1)

with open("VERSION") as f:
    VERSION = f.read()

with open("requirements.txt") as f:
    requirements = [line.strip() for line in f if line.strip() != ""]

with open("README.md") as f:
    long_description = f.read()

data_files = [
    ("symfem", ["symfem/VERSION"])]


if __name__ == "__main__":
    setuptools.setup(
        name="symfem",
        description="a symbolic finite element definition library",
        long_description=long_description,
        long_description_content_type="text/markdown",
        version=VERSION,
        author="Matthew Scroggs",
        license="MIT",
        author_email="symfem@mscroggs.co.uk",
        maintainer_email="symfem@mscroggs.co.uk",
        url="https://github.com/mscroggs/symfem",
        packages=["symfem", "symfem.core", "symfem.elements"],
        data_files=data_files,
        include_package_data=True,
        install_requires=requirements
    )



import pytest
import sympy
from symfem import create_element
from symfem.core.symbolic import subs, x, PiecewiseFunction
from symfem.core.vectors import vsub
from utils import test_elements, all_symequal


@pytest.mark.parametrize(
    ("cell_type", "element_type", "order", "variant"),
    [[reference, element, order, variant]
     for reference, i in test_elements.items() for element, j in i.items()
     for variant, k in j.items() for order in k])
def test_element_functionals_and_continuity(elements_to_test, cells_to_test,
                                            cell_type, element_type, order, variant):
    if elements_to_test != "ALL" and element_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell_type not in cells_to_test:
        pytest.skip()

    # Test functionals
    space = create_element(cell_type, element_type, order, variant)
    for i, f in enumerate(space.get_basis_functions()):
        for j, d in enumerate(space.dofs):
            if i == j:
                assert d.eval(f).expand().simplify() == 1
            else:
                assert d.eval(f).expand().simplify() == 0
            assert d.entity_dim() is not None

    if order > 4:
        return  # For high order, testing continuity is slow

    # Test continuity
    if cell_type == "interval":
        vertices = ((-1, ), (0, ))
        entity_pairs = [[0, (0, 1)]]
    elif cell_type == "triangle":
        vertices = ((-1, 2), (0, 0), (0, 1))
        entity_pairs = [[0, (0, 1)], [0, (2, 2)], [1, (1, 0)]]
    elif cell_type == "tetrahedron":
        vertices = ((-1, 0, 0), (0, 0, 0), (0, 1, 0), (0, 0, 1))
        entity_pairs = [[0, (0, 1)], [0, (2, 2)], [0, (3, 3)],
                        [1, (0, 0)], [1, (3, 1)], [1, (4, 2)],
                        [2, (1, 0)]]
    elif cell_type == "quadrilateral":
        vertices = ((0, 0), (0, 1), (-1, 0), (-1, 1))
        entity_pairs = [[0, (0, 0)], [0, (2, 1)], [1, (1, 0)]]
    elif cell_type == "hexahedron":
        vertices = ((0, 0, 0), (0, 0, 1), (0, 1, 0), (0, 1, 1),
                    (-1, 0, 0), (-1, 0, 1), (-1, 1, 0), (-1, 1, 1))
        entity_pairs = [[0, (0, 0)], [0, (2, 2)], [0, (4, 1)], [0, (6, 3)],
                        [1, (1, 1)], [1, (2, 0)], [1, (6, 5)], [1, (9, 3)],
                        [2, (0, 2)]]

    if space.continuity == "L2":
        return

    for dim, entities in entity_pairs:
        for fi, gi in zip(*[space.entity_dofs(dim, i) for i in entities]):
            basis = space.get_basis_functions()
            basis2 = space.map_to_cell(vertices)
            f = basis[fi]
            g = basis2[gi]

            if isinstance(f, PiecewiseFunction):
                assert space.reference.tdim == 2
                f = f.get_piece((0, sympy.Rational(1, 2)))
                g = g.get_piece((0, sympy.Rational(1, 2)))

            f = subs(f, [x[0]], [0])
            g = subs(g, [x[0]], [0])

            if space.continuity == "C0":
                pass
            elif space.continuity == "C1":
                f = [f] + [f.diff(i) for i in x[:space.reference.tdim]]
                g = [g] + [g.diff(i) for i in x[:space.reference.tdim]]
            elif space.continuity == "H(div)":
                f = f[0]
                g = g[0]
            elif space.continuity == "H(curl)":
                f = f[1:]
                g = g[1:]
            elif space.continuity == "inner H(curl)":
                if len(vertices[0]) == 2:
                    f = f[3]
                    g = g[3]
                if len(vertices[0]) == 3:
                    if dim == 1:
                        vs = space.reference.sub_entities(1)[entities[0]]
                        v0 = space.reference.vertices[vs[0]]
                        v1 = space.reference.vertices[vs[1]]
                        tangent = vsub(v1, v0)
                        f = sum(i * f[ni * len(tangent) + nj] * j
                                for ni, i in enumerate(tangent)
                                for nj, j in enumerate(tangent))
                        g = sum(i * g[ni * len(tangent) + nj] * j
                                for ni, i in enumerate(tangent)
                                for nj, j in enumerate(tangent))
                    else:
                        assert dim == 2
                        f = [f[4], f[8]]
                        g = [g[4], g[8]]
            elif space.continuity == "inner H(div)":
                f = f[0]
                g = g[0]
            elif space.continuity == "integral inner H(div)":
                f = f[0].integrate((x[1], 0, 1))
                g = g[0].integrate((x[1], 0, 1))
            else:
                raise ValueError(f"Unknown continuity: {space.continuity}")

            assert all_symequal(f, g)


@pytest.mark.parametrize("n_tri", [3, 4, 6, 8])
@pytest.mark.parametrize("order", range(2))
def test_dual_elements(n_tri, order):
    space = create_element(f"dual polygon({n_tri})", "dual", order)

    sub_e = create_element("triangle", space.fine_space, space.order)
    for f, coeff_list in zip(space.get_basis_functions(), space.dual_coefficients):
        for piece, coeffs in zip(f.pieces, coeff_list):
            map = sub_e.reference.get_map_to(piece[0])
            for dof, value in zip(sub_e.dofs, coeffs):
                point = subs(map, x, dof.point)
                assert all_symequal(value, subs(piece[1], x, point))


@pytest.mark.parametrize("n_tri", [3, 4])
@pytest.mark.parametrize("element_type", ["BC", "RBC"])
def test_bc_elements(n_tri, element_type):
    create_element(f"dual polygon({n_tri})", element_type, 1)



"""Mardal-Tai-Winther elements on simplices.

This element's definition appears in https://doi.org/10.1137/S0036142901383910
(Mardal, Tai, Winther, 2002)
and https://doi.org/10.1007/s10092-006-0124-6 (Tail, Mardal, 2006)
"""

from ..core.finite_element import FiniteElement
from ..core.moments import make_integral_moment_dofs
from ..core.polynomials import polynomial_set
from ..core.symbolic import x, zero, one
from ..core.calculus import curl
from ..core.functionals import NormalIntegralMoment, TangentIntegralMoment, IntegralMoment
from .lagrange import DiscontinuousLagrange
from .nedelec import NedelecFirstKind


class MardalTaiWinther(FiniteElement):
    """Mardal-Tai-Winther Hdiv finite element."""

    def __init__(self, reference, order, variant):
        assert order == 3

        dofs = make_integral_moment_dofs(
            reference, facets=(NormalIntegralMoment, DiscontinuousLagrange, 1),
            variant=variant)

        if reference.name == "triangle":
            poly = [(one, zero), (x[0], zero), (x[1], zero),
                    (zero, one), (zero, x[0]), (zero, x[1]),
                    # (x**2 + 2*x*y, -2*x*y - y**2)
                    (x[0] ** 2 + 2 * x[0] * x[1],
                     -2 * x[0] * x[1] - x[1] ** 2),
                    # (-x**3 + 2*x**2 + 3*x*y**2, 3*x**2*y - 4*x*y - y**3)
                    (-x[0] ** 3 + 2 * x[0] ** 2 + 3 * x[0] * x[1] ** 2,
                     3 * x[0] ** 2 * x[1] - 4 * x[0] * x[1] - x[1] ** 3),
                    # (2*x**2*y + x**2 + 3*x*y**2, -2*x*y**2 - 2*x*y - y**3)
                    (2 * x[0] ** 2 * x[1] + x[0] ** 2 + 3 * x[0] * x[1] ** 2,
                     -2 * x[0] * x[1] ** 2 - 2 * x[0] * x[1] - x[1] ** 3)]
            dofs += make_integral_moment_dofs(
                reference, facets=(TangentIntegralMoment, DiscontinuousLagrange, 0),
                variant=variant)
        else:
            assert reference.name == "tetrahedron"

            poly = polynomial_set(reference.tdim, reference.tdim, 1)
            for p in polynomial_set(reference.tdim, reference.tdim, 1):
                poly.append(curl(tuple(i * x[0] * x[1] * x[2] * (1 - x[0] - x[1] - x[2])
                                       for i in p)))

            dofs += make_integral_moment_dofs(
                reference, facets=(IntegralMoment, NedelecFirstKind, 1),
                variant=variant)

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["Mardal-Tai-Winther", "MTW"]
    references = ["triangle", "tetrahedron"]
    min_order = 3
    max_order = 3
    mapping = "contravariant"
    continuity = "H(div)"
